#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""These functions provide some mechanisms for dealing with
ISIS cube files.  These functions are not comprehensive, and only
seek to provide functionality that does not exist elsewhere.

For example, there is already a GDAL driver for ISIS cubes, and
access to the primary bands and such can already be accomplished
via GDAL, and in order to get the image pixels as a numpy array::

    import numpy as np
    from osgeo import gdal_array

    cube = 'some.cub'
    img_arr = gdal_array.LoadFile(cube)

If you want to make sure to mask out all of the special pixels
in the image you've read into img_arr above, you can do this::

    import pvl
    import kalasiris as isis

    label = pvl.load(cube)
    specialpix = getattr(isis.specialpixels,
                         label['IsisCube']['Core']['Pixels']['Type'])
    masked_img_arr = np.ma.masked_outside(img_arr,
                                          specialpix.Min, specialpix.Max)

"""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import os
import struct
from collections import abc
from warnings import warn

from .kalasiris import getkey

data_sizes = {'Integer': 4,
              'Double': 8,
              'Real': 4,
              'Text': 1}
data_formats = {'Integer': 'i',
                'Double': 'd',
                'Real': 'f'}


def _get_start_size(d: dict) -> tuple((int, int)):
    """Returns a tuple of ints that represent the true start byte and size
    based on the provided dict.

    This is a convenience function such that if you have these values in a
    dictionary with a *StartByte* and *Bytes* keys.

    Since the StartByte labels on an ISIS cube are 1-based counts, and not
    0-based, in order to convert from the numbers provided in the labels
    to a byte count for Python's read() functions, you must subtract one
    from the listed StartByte, which this function does.
    """
    start = int(d['StartByte']) - 1
    size = int(d['Bytes'])
    return (start, size)


def get_startsize_from(label=None, table_name=None,
                       cube_path=None) -> tuple((int, int)):
    """Returns a tuple of ints that represent the true start byte and size
    based on the provided *label* or combination of *table_name* and
    *cube_path*.

    Either *label* or *table_name* and *cube_path* are needed.  If
    neither a *label* nor a *table_name* is provided, then this
    function will raise a ValueError.  If both are provided, *label*
    will take precedence and *table_name* will be ignored.

    *label* is a dict which must contain a *StartByte* key and a *Bytes*
    key whose values can be converted to int (if not already).  These
    values should be those in the cube file label for the table.

    The name of the table as a string can be provided via *table_name*
    and the ISIS getkey function will be applied to the file at
    *cube_path* to extract the needed StartByte and Bytes values
    from the label.  However, if there is more than one table in
    the cube, getkey can only find the first, and a ValueError might
    be returned.

    If the pvl library is available, this function will use it to find
    *all* of the tables in the *cube_path* labels and will find the one
    named by *table_name* if it is present.
    """
    if label is None and table_name is None:
        raise ValueError("Neither label nor table_name were provided.")

    if label is not None:
        return _get_start_size(label)
    else:
        try:
            import pvl
            label = pvl.load(str(cube_path))
            for t in label.getlist('Table'):
                if t['Name'] == table_name:
                    return _get_start_size(t)
            else:
                raise KeyError(f"There is no table '{table_name}' in the "
                               f"labels of {cube_path}")
        except ImportError:
            name = getkey(cube_path, objname='Table',
                          keyword='Name').stdout.strip()
            if table_name == name:
                return _get_start_size(
                    {'StartByte':
                     getkey(cube_path, objname='Table',
                            keyword='StartByte').stdout.strip(),
                     'Bytes':
                     getkey(cube_path, objname='Table',
                            keyword='Bytes').stdout.strip()})
            else:
                raise KeyError(f"The first table in {cube_path} that "
                               "ISIS getkey could find is not named "
                               f"{table_name} it is {name}.  If your "
                               "cube has more than one table, you can "
                               "provide the info via the label dict "
                               "instead of table_name, or if you install "
                               "the pvl Python library then this function "
                               "can use it to find all of the tables.")


# This function is derived from this commit dated Sep 24, 2019:
# https://github.com/USGS-Astrogeology/ale/commit/add5368ba46b2c911de9515afeaccc4d1c981000
def read_table_data(cube_path: os.PathLike,
                    label=None, table_name=None) -> bytes:
    """Returns a bytes object with the contents read from the file at
    *cube_path* based on the elements provided in the *label* or
    *table_name*.

    Either *label* or *table_name* is needed.  If neither a *label*
    nor a *table_name* is provided, then this function will raise
    a ValueError.  If both are provided, *label* will take precedence
    and *table_name* will be ignored.

    *label* is a dict which must contain a *StartByte* key and a *Bytes*
    key whose values can be converted to int (if not already).  These
    values should be those in the cube file label for the table.

    The name of the table as a string can be provided via *table_name*
    and the ISIS getkey function will be applied to extract the needed
    StartByte and Bytes values from the label.  However, if there is more
    than one table in the cube, getkey can only find the first, and a
    ValueError might be returned.

    If the pvl library is available, this function will use it to find
    *all* of the tables in the *cube_path* labels and will find the one
    named by *table_name* if it is present.
    """

    (start, size) = get_startsize_from(label, table_name, cube_path)

    with open(cube_path, "rb") as cubehandle:
        cubehandle.seek(start)
        table = cubehandle.read(size)

    return table


# This function is derived from this commit dated Sep 24, 2019:
# https://github.com/USGS-Astrogeology/ale/commit/add5368ba46b2c911de9515afeaccc4d1c981000
def parse_table(data: bytes, fields: list) -> dict:
    """Return a Python dictionary created from the bytes *data* of
    an ISIS cube table (presumably extracted via read_table_data()),
    and described by the *fields* list and *records*.

    Please be aware that this does not perform masking of the ISIS
    special pixels that may be present in the table, and simply
    returns them as the appropriate int or float values.

    The *fields* list must be a list of dicts, each of which must
    contain the following keys: 'Name', 'Type', and 'Size'.  The
    'Name' key can be any string (and these will end up being the
    keys in the returned dict).  'Size' is the size in bytes of the
    field, and 'Type' is a string that must be one of 'Integer',
    'Double', 'Real', or 'Text'.

    If you are using the pvl library, the get_table() function will
    be easier to use.
    """

    row_len = 0
    for f in fields:
        row_len += data_sizes[f['Type']] * int(f['Size'])
    if len(data) % row_len != 0:
        raise ValueError(f"The total sizes of each field ({row_len}) do not "
                         "evenly divide into the size of the data "
                         f"({len(data)}), so something is off.")

    # Parse the binary data
    results = {f['Name']: [] for f in fields}
    offset = 0
    while offset < len(data):
        for f in fields:
            if f['Type'] == 'Text':
                field_data = data[offset:offset + int(f['Size'])].decode(
                    encoding='latin_1')
            else:
                data_fmt = data_formats[f['Type']] * int(f['Size'])
                f_data = struct.unpack_from(data_fmt, data, offset)
                if len(f_data) == 1:
                    field_data = f_data[0]
                else:
                    field_data = list(f_data)

            results[f['Name']].append(field_data)
            offset += data_sizes[f['Type']] * int(f['Size'])

    return results


# This function is derived from this commit dated Sep 24, 2019:
# https://github.com/USGS-Astrogeology/ale/commit/add5368ba46b2c911de9515afeaccc4d1c981000
def get_table(cube_path: os.PathLike, table_name: str) -> dict:
    """Return a Python dictionary created from the named table in
    the ISIS cube.

    This function requires the pvl Python library.
    """
    # Toyed with allowing a file_object=None argument, docstring would have
    # been:
    #
    # If the optional *file_object* is given, it should be the result
    # of opening *cube_path* which is readable.  This simply allows
    # a caller to provide an already-opened file object.  Otherwise,
    # this function will open and then close the file at *cube_path*.
    #
    # Decided against it, the potential to pass a file_object that was
    # *different* from an opened *cube_path* had the potential for much
    # mayhem without a tremendous amount of gain.

    try:
        import pvl
        label = pvl.load(str(cube_path))
        table_label = None
        for t in label.getlist('Table'):
            if t['Name'] == table_name:
                table_label = t
                break

        # if file_object is not None:
        #     (start, size) = _get_start_size(table_label)
        #     file_object.seek(start)
        #     table_data = file_object.read(size)
        # else:
        table_data = read_table_data(cube_path, table_label)
        return parse_table(table_data, table_label.getlist('Field'))

        # The original ale function added the keywords into the returned
        # table, but that doesn't seem like a great idea, since that means
        # that those keys are 'special' meta-data keys, whereas the other
        # keys in the returned dict are 'regular' field keys, and once
        # returned, there's no way to know which is which.
        # # Add the keywords from the label
        # results.update({
        #     key: value for key,
        #                    value in table_label.items() if not isinstance(
        #         value, pvl._collections.PVLGroup)})
        #
        # return results

    except ImportError:
        warn("The pvl library is not present, so get_table() cannot be used. "
             "The parse_table() function might work for you.", ImportWarning)
        raise


def overwrite_table_data(cube_path: os.PathLike, data: bytes,
                         label=None, table_name=None):
    """The file at *cube_path* will be modified by overwriting the
    data in the specfied table name with the contents of *data*.

    Either *label* or *table_name* is needed.  If neither a *label*
    dict (which must contain a 'Name' key) nor a *table_name* is
    provided, then this function will raise a ValueError.  If both
    are provided, *label* will take precedence and *table_name*
    will be ignored.

    *label* is a dict which must contain *Name*, *StartByte*, and
    *Bytes* keys (*StartByte* and *Bytes* must be convertable to
    int if not already).  These values will be used to locate where
    in the file to write the new *data*.

    The name of the table as a string can be provided via *table_name*
    and the ISIS getkey function will be applied to extract the needed
    StartByte and Bytes values from the label.  However, if there is more
    than one table in the cube, getkey can only find the first, and a
    ValueError might be returned, even thought there is a table of that
    name in the file.

    If the pvl library is available, this function will use it to find
    *all* of the tables in the *cube_path* labels and will find the one
    named by *table_name* if it is present.
    """

    (start, size) = get_startsize_from(label, table_name, cube_path)

    if size != len(data):
        raise ValueError(f"The size of the table ({size}) to be overwritten "
                         f"from the file  ({cube_path}) is different from "
                         f"size of the data provided ({len(data)}).")

    with open(cube_path, "r+b") as cubehandle:
        cubehandle.seek(start)
        cubehandle.write(data)

    return


def encode_table(table: dict, fields: list) -> bytes:
    """Return a bytes object created from the *table* dict.

    The *table* dict must contain lists of equal length as values.
    If they are not of equal length, an IndexError will be raised.

    The *fields* list must be a list of dicts, each of which must
    contain the following keys: 'Name', 'Type', and 'Size'.  The
    'Name' key can be any string, but must match the keys in the
    *table* dict.  'Size' is the size in bytes of the field, and
    'Type' is a string that must be one of 'Integer', 'Double',
    'Real', or 'Text'.

    If a field's 'Size' value is more than 1, then the list which
    is the value of the *table* with that key name must be a list of
    length 'Size'.

    If you are using the pvl library, the overwrite_table() function will
    be easier to use.
    """

    field_lengths = set()
    for v in table.values():
        field_lengths.add(len(v))

    if not len(field_lengths) == 1:
        raise IndexError("At least one of the lists in the table has "
                         f"a different length than the rest: {field_lengths}")

    data = bytes()
    for row in range(field_lengths.pop()):
        for f in fields:
            obj = table[f['Name']][row]
            size = int(f['Size'])
            if f['Type'] == 'Text':
                if len(obj) > size:
                    raise IndexError(f"The length of {obj} ({len(obj)}) is "
                                     "larger than the allowable Size of the "
                                     f"field ({size})")
                else:
                    data += obj.ljust(size).encode(encoding="latin_1")
            else:
                data_fmt = data_formats[f['Type']] * size
                if isinstance(obj, abc.Sequence):
                    if len(obj) == size:
                        data += struct.pack(data_fmt, *obj)
                    else:
                        raise IndexError(f"The length of {obj} ({len(obj)}) "
                                         " is different than the Size of the "
                                         f"field ({size})")
                elif size == 1:
                    data += struct.pack(data_fmt, obj)
                else:
                    raise ValueError(f"There is only a single value ({obj}) "
                                     "but the field indicates there should "
                                     f"be {int(f['Size'])}.")

    return data


def overwrite_table(cube_path: os.PathLike, table_name: str, table: dict):
    """The file at *cube_path* will be modified by overwriting the
    data in the specfied table name with the contents of *table*.

    The *table* dict must contain lists of equal length as values.
    If they are not of equal length, an IndexError will be raised.
    The *table* dict must also contain, as keys, all of the Field names
    from *table_name* in the *cube_path*.

    This function requires the pvl Python library.
    """

    try:
        import pvl
        label = pvl.load(str(cube_path))
        table_label = None
        for t in label.getlist('Table'):
            if t['Name'] == table_name:
                table_label = t
                break

        data = encode_table(table, table_label.getlist('Field'))
        overwrite_table_data(cube_path, data, table_label)

        return

    except ImportError:
        warn("The pvl library is not present, so overwrite_table() cannot be "
             "used. The overwrite_table_data() function might work for you.",
             ImportWarning)
        raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the pysis emulation module."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import contextlib
# import subprocess
import unittest
from pathlib import Path

import kalasiris.pysis as pysis
from .utils import resource_check as rc


run_real_files = True
run_real_files_reason = 'Tests on real files, and runs ISIS.'

# Hardcoding this, but I sure would like a better solution.
HiRISE_img = Path('test-resources') / 'PSP_010502_2090_RED5_0.img'
img = HiRISE_img


@unittest.skipUnless(run_real_files, run_real_files_reason)
class TestResources(unittest.TestCase):
    '''Establishes that the test image exists.'''

    def test_resources(self):
        (truth, test) = rc(img)
        self.assertEqual(truth, test)


class Test_get_isis_program_names(unittest.TestCase):

    @unittest.skipUnless(run_real_files, run_real_files_reason)
    def test_get_names(self):
        self.assertIn('cam2map', dir(pysis))


@unittest.skipUnless(run_real_files, run_real_files_reason)
class Test_hi2isis(unittest.TestCase):

    def setUp(self):
        self.img = img

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()

    def test_hi2isis(self):
        tocube = Path('test_hi2isis.cub')
        pysis.hi2isis(self.img, to=tocube)
        self.assertTrue(tocube.is_file())
        tocube.unlink()


@unittest.skipUnless(run_real_files, run_real_files_reason)
class Test_getkey(unittest.TestCase):

    def setUp(self):
        self.cub = Path('test_getkey.cub')
        pysis.hi2isis(img, to=self.cub)

    def tearDown(self):
        self.cub.unlink()
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()

    def test_getkey(self):
        truth = b'HIRISE\n'
        key = pysis.getkey(self.cub, grpname='Instrument',
                           keyword='InstrumentId')
        self.assertEqual(truth, key)

    def test_getkey_fail(self):
        # Pixels doesn't have InstrumentId, should fail
        self.assertRaises(pysis.ProcessError,
                          pysis.getkey, self.cub,
                          grpname='Pixels', keyword='InstrumentId')

    def test_getkey_k_fail(self):
        # Calling getkey with getkey_k syntax will fail
        self.assertRaises(IndexError,
                          pysis.getkey, self.cub,
                          'Instrument', 'InstrumentId')



#!/usr/bin/env python
"""This module contains the PathSet Class.

orking with ISIS can result in a lot of files to keep track of.
he PathSet Class is simply a mutable set that only takes
class:`pathlib.Path` objects.  If you need to keep track of a
unch of files (typically to delete them after a set of processing
alls), you can use a :class:`.PathSet` to keep track of them, and then
elete them, like so::

    import kalasiris as isis

    to_delete = isis.PathSet()

    input_p = Path('some.fits')
    output_p = input_p.with_suffix('.done.cub')

    isis_cub = to_delete.add(input_p.with_suffix('.cub'))
    isis.lorri2isis(input_fits, to= isis_cub)

    first = to_delete.add(input_p..with_suffix('.1.cub'))
    isis.some_progeram(isis_cub, to=first)

    second = to_delete.add(input_p..with_suffix('.2.cub'))
    isis.some_program(first, to=second)

    isis.final_step(second, to=output_p)

    to_delete.unlink()
"""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from pathlib import Path


class PathSet(set):
    """A class for containing a set of :class:`pathlib.Path` objects."""

    def __init__(self, iterable=None):
        if iterable:
            for value in iterable:
                if not isinstance(value, Path):
                    raise TypeError('only accepts pathlib.Path objects')
            super().__init__(iterable)
        else:
            super().__init__()

    def add(self, elem):
        """This variation on add() returns the element."""
        if not isinstance(elem, Path):
            raise TypeError('only accepts pathlib.Path objects')
        if elem in self:
            raise ValueError(f'The {elem} object is already a '
                             'member of the PathSet.')
        super().add(elem)
        return elem

    def unlink(self):
        """Just runs Path.unlink() on all members."""
        for p in self:
            p.unlink()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `utils` module."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from pathlib import Path
import unittest
from .utils import resource_check as rc


class TestResources(unittest.TestCase):

    def setUp(self):
        t = Path('tests')
        self.exists1 = t / 'test_kalasiris.py'
        self.exists2 = t / 'test_utils.py'
        self.not1 = t / 'notexists' / 'foo'
        self.not2 = t / 'notexists' / 'foo2'

    def test_have(self):
        (truth, test) = rc(self.exists1, self.exists2)
        self.assertEqual(truth, test)

    def test_have_none(self):
        (truth, test) = rc(self.not1, self.not2)
        self.assertNotEqual(truth, test)

    def test_have_some(self):
        (truth, test) = rc(self.exists1, self.not1)
        self.assertNotEqual(truth, test)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `cubenormDialect` class."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import contextlib
import csv
import io
import unittest
from unittest.mock import call, Mock
from pathlib import Path

import kalasiris as isis
from .utils import resource_check as rc


run_real_files = True
run_real_files_reason = 'Tests on real files, and runs ISIS.'

# Hardcoding this, but I sure would like a better solution.
img = Path('test-resources') / 'PSP_010502_2090_RED5_0.img'


@unittest.skipUnless(run_real_files, run_real_files_reason)
class TestResources(unittest.TestCase):
    '''Establishes that the test image exists.'''

    def test_resources(self):
        (truth, test) = rc(img)
        self.assertEqual(truth, test)


class TestCubenormFile(unittest.TestCase):

    def setUp(self):
        self.stats = (
            '    Band  RowCol    ValidPoints        Average         '
            'Median         StdDev        Minimum        Maximum\n'
            '       1       1           4000        6532.67           '
            '6526         191.58           5790           7211\n'
            '       1       2           4000        6296.87           '
            '6303        187.865           5501           6980\n'
            '       1       3           4000        6341.42           '
            '6378        188.321           5465           7026\n'
            '       1       4           4000        6381.09           '
            '6415        187.721           5429           6980\n'
            '       1       5           4000        6391.85           '
            '6415        184.816           5537           7026\n'
            '       1       6           4000        6411.03           '
            '6415        184.951           5465           7073\n')

    def test_Dialect(self):
        reader = csv.reader(self.stats.splitlines(),
                            dialect=isis.cubenormfile.Dialect)
        for row in reader:
            self.assertEqual(8, len(row))
            break

    def test_writer(self):
        columns = list()
        reader = csv.reader(self.stats.splitlines(),
                            dialect=isis.cubenormfile.Dialect)
        for row in reader:
            columns.append(row)

        csvfile = Mock()
        writer = isis.cubenormfile.writer(csvfile)
        for row in columns:
            writer.writerow(row)

        expected = list()
        for line in self.stats.splitlines(True):
            expected.append(call.write(line))

        self.assertEqual(csvfile.method_calls, expected)

    def test_DictWriter(self):
        columns = list()
        reader = csv.DictReader(self.stats.splitlines(),
                                dialect=isis.cubenormfile.Dialect)
        for row in reader:
            columns.append(row)

        csvfile = Mock()
        writer = isis.cubenormfile.DictWriter(csvfile)
        writer.writeheader()
        for row in columns:
            writer.writerow(row)

        expected = list()
        for line in self.stats.splitlines(True):
            expected.append(call.write(line))

        self.assertEqual(csvfile.method_calls, expected)


@unittest.skipUnless(run_real_files, run_real_files_reason)
class TestCubenormFile_filesystem(unittest.TestCase):

    def setUp(self):
        self.cube = Path('test_cubenormfile.cub')
        self.statsfile = Path('test_cubenormfile.stats')
        isis.hi2isis(img, to=self.cube)
        isis.cubenorm(self.cube, stats=self.statsfile)

    def tearDown(self):
        with contextlib.suppress(FileNotFoundError):
            Path('print.prt').unlink()
        self.cube.unlink()
        self.statsfile.unlink()

    def test_Dialect(self):
        with open(self.statsfile) as csvfile:
            reader = csv.reader(csvfile, dialect=isis.cubenormfile.Dialect)
            for row in reader:
                self.assertEqual(8, len(row))
                break

    def test_writer(self):
        columns = list()
        with open(self.statsfile) as csvfile:
            reader = csv.reader(csvfile, dialect=isis.cubenormfile.Dialect)
            for row in reader:
                columns.append(row)

        outfile = self.statsfile.with_suffix('.writer.stats')
        with open(outfile, 'w') as csvfile:
            writer = isis.cubenormfile.writer(csvfile)
            for row in columns:
                writer.writerow(row)

        with open(self.statsfile) as in_file:
            with open(outfile) as out_file:
                for i, o in zip(in_file.readlines(), out_file.readlines()):
                    with self.subTest(inputline=i, outputline=o):
                        self.assertEqual(i, o)
        outfile.unlink()

    def test_DictWriter(self):
        columns = list()
        with open(self.statsfile) as csvfile:
            reader = csv.DictReader(csvfile, dialect=isis.cubenormfile.Dialect)
            for row in reader:
                columns.append(row)

        with io.StringIO() as out_file:
            writer = isis.cubenormfile.DictWriter(out_file)
            writer.writeheader()
            for row in columns:
                writer.writerow(row)

            with open(self.statsfile) as in_file:
                for i, o in zip(in_file.readlines(), out_file.readlines()):
                    with self.subTest(inputline=i, outputline=o):
                        self.assertEqual(i, o)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Provides the ability to call ISIS functions."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

# Thou shalt only import from the Python Standard Library.
import os
import subprocess
import sys
from pathlib import Path
# This file shall have *NO* non-Standard Library dependencies.

# kalasiris library version:
__version__ = '1.7.0'

# These definitions and the use of env= in the subprocess.run calls allow us to
# run ISIS in a very lean environment.  Of course, users can override with
# their complete environment by making kalasiris.environ = os.environ
# before any calls to ISIS programs.
_isisroot = os.environ['ISISROOT']
_isis3data = os.environ['ISIS3DATA']
environ = {'ISISROOT':  _isisroot,
           'ISIS3DATA': _isis3data,
           'PATH':      str(Path(_isisroot) / 'bin'),
           'HOME':      os.path.expanduser('~')}
# If we don't also set $HOME, ISIS tries to make a local ./\$HOME dir
# Can't just use os.environ['HOME'] because not all platforms have
# that environment variable set (Windows uses something different).

# These are the names of the reserved parameters that can be given
# as arguments to any ISIS program, prefixed by a dash (-).
_res_param_no_vals = set(['webhelp', 'last', 'gui', 'nogui', 'verbose'])
_res_param_maybe = set(['help', 'log', 'info', 'save'])

# The ISIS programs in this list do not follow the 'normal' argument
# patters for most ISIS programs, they just consume everything you
# give them, so we need to treat them differently.
_pass_through_programs = set(['cneteditor', 'qmos', 'qnet', 'qtie', 'qview'])


def param_fmt(key: str, value: str) -> str:
    """Returns a "key=value" string from the inputs.

    This is the pattern that ISIS uses for arguments.  If there are
    any trailing underbars (_) on the key, they will be stripped
    off.  This supports the old pysis syntax to protect Python
    reserved words like from and min, while still allowing the user
    to provide 'natural' function calls like isis.stats(from_=cubefile).

    Additionally, it also supports passing what ISIS calls *reserved
    parameters* for any ISIS program, denoted with a prefix of a single
    dash, like ``-restore=file`` or ``-verbose`` via keys with two trailing
    underbars.  So to call ``spiceinit from= some.cub -restore=file`` you
    would do this::

        cubefile = 'some.cub'
        restore_file = 'some.par'
        isis.spiceinit(cubefile, restore__=restore_file)

    Likewise, to call ``getkey -help``, or ``getkey -help=GRPNAME`` do this::

        isis.getkey('help__')
        isis.getkey(help__='GRPNAME')

    Of course, you'll probably want to do this::

        help_text = isis.getkey(help__='').stdout
    """
    # The logic for dealing with a single parameter, like
    # isis.getkey('help__') # is down in the _build_isis_fn() factory
    # method.
    if key.endswith('__') and key.rstrip('_') in _res_param_maybe:
        return '-{}={}'.format(key.rstrip('_'), value)
    else:
        return '{}={}'.format(key.rstrip('_'), value)


def _run_isis_program(cmd: list) -> subprocess.CompletedProcess:
    """Wrapper for subprocess.run()"""
    return subprocess.run(cmd, env=environ, check=True,
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                          universal_newlines=True)


def _build_isis_fn(fn_name: str):
    """This factory builds a simple function to call an ISIS program."""

    # Define the structure of the generic function, isis_fn:
    def isis_fn(*args, **kwargs) -> subprocess.CompletedProcess:
        __name__ = fn_name                 # noqa: F841
        __doc__ = f'Runs ISIS3 {fn_name}'  # noqa: F841
        cmd = [fn_name]
        if fn_name in _pass_through_programs:
            cmd.extend(args)
        else:
            args_list = list(args)
            if len(args) > 0 and not str(args[0]).endswith('__'):
                cmd.append(param_fmt('from', args_list.pop(0)))
            for a in args_list:
                if a.endswith('__') and a.rstrip('_') in \
                        _res_param_no_vals.union(_res_param_maybe):
                    cmd.append('-{}'.format(a.rstrip('_')))
                else:
                    e = ('only accepts 1 non-keyword argument '
                         '(and sets it to from= ) '
                         'not sure what to do with ' + a)
                    raise IndexError(e)
            cmd.extend(map(param_fmt, kwargs.keys(), kwargs.values()))
        return(_run_isis_program(cmd))

    # Then add it, by name to the enclosing module.
    setattr(sys.modules[__name__], fn_name, isis_fn)
    # Could have also used sys.modules['kalasiris'] if I wanted to be explicit.


def _get_isis_program_names():
    """Returns an iterable of ISIS program names.

    With the new conda distribution, there is a lot of stuff in
    $ISISROOT/bin that isn't actually an ISIS program.  So just
    slurping the names in $ISISROOT/bin gets too many things that
    aren't ISIS programs.

    Instead, the isis conda distribution provides $ISISROOT/bin/xml/
    which contains the documentation XML files.  Every XML file
    in that directory corresponds to the name of an ISIS program,
    which is perfect.
    """
    bindir = Path(environ['ISISROOT']) / 'bin'
    xmldir = bindir / 'xml'
    for entry in xmldir.iterdir():
        if(entry.is_file() and
           (bindir / entry.stem).is_file() and
           os.access(bindir / entry.stem, os.X_OK) and
           not entry.name.startswith('.')):
            if '.xml' == entry.suffix:
                yield entry.stem


# Now use the builder function to automatically create functions
# with these names:
for p in _get_isis_program_names():
    _build_isis_fn(p)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Provides kalasiris *_k functions*.

The kalasiris *_k functions* provide some syntactic sugar to make
calling the ISIS programs just that much easier.  For example here
are two ways to do the same thing::

     import kalasiris as isis

     cube_file = 'some.cub'

     keyval = isis.getkey(cube_file, grpname='Instrument',
                      keyword='InstrumentId').stdout.strip()

     k_keyval = isis.getkey_k(cube_file, 'Instrument', 'InstrumentId')

And the values of ``keyval`` and ``k_keyval`` are identical, its
just that the *_k function* version is a little more compact.
Each of the *_k functions* implements their modifications a
little differently, so make sure to read their documentation.
"""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import tempfile
import os
from pathlib import Path

import kalasiris as isis


def getkey_k(cube: os.PathLike, group: str, key: str) -> str:
    """Simplified calling for getkey.

    No default parameters are needed, and it directly returns a string.
    """
    return(isis.getkey(cube, grpname=group, keyword=key).stdout.strip())


def hi2isis_k(*args, **kwargs):
    """Creates a default name for the to= cube.

    If the FROM file has the name ``foo.img``, then the output will
    be ``foo.cub``.
    """
    if not any(k.startswith('t') for k in kwargs.keys()):
        from_path = Path()
        if len(args) > 0 and not str(args[0]).endswith('__'):
            from_path = Path(args[0])
        else:
            for(k, v) in kwargs.items():
                if k.startswith('f'):
                    from_path = Path(v)
                    break

        kwargs.setdefault('to', from_path.with_suffix('.cub'))

    return(isis.hi2isis(*args, **kwargs))


def hist_k(*args, **kwargs) -> str:
    """Returns the contents of the file created by ISIS hist as a string.

    If there is a TO= parameter in the arguments, ``hist_k()`` will
    create the file, and return its contents as a string
    """
    to_pathlike = None
    for (k, v) in kwargs.items():
        if 'to' == k or 'to_' == k:
            to_pathlike = v

    f = None
    if not to_pathlike:
        f = tempfile.NamedTemporaryFile(mode='w+')
        kwargs['to'] = f.name

    isis.hist(*args, **kwargs)

    if not f:
        f = open(to_pathlike, 'r')
    contents = f.read()
    f.close()

    return contents


def cubeit_k(fromlist: list, **kwargs):
    """Takes a list of paths to cubes to operate cubeit on,
    rather than having the user create a text list.
    """
    with isis.fromlist.temp(fromlist) as f:
        kwargs['fromlist'] = f
        cp = isis.cubeit(**kwargs)

    return cp


def stats_k(*args, **kwargs) -> dict:
    """Returns the result of running ISIS stats as a Python Dictionary.

    If there are TO=, FORMAT=, or APPEND= parameters, ``stats_k`` will
    perform the file-based activities that ``stats`` normally would,
    and also return the Python Dictionary.
    """
    # We could use the pvl library to parse the returned text, but
    # that would involve a dependency, and since the format is so
    # simple, we'll just parse it directly here.

    stats_text = isis.stats(*args, **kwargs).stdout

    d = dict()
    for line in filter(lambda x: '=' in x, stats_text.splitlines()):
        (k, equals, v) = line.partition('=')
        if 'Group' in k:
            continue
        d[k.strip()] = v.strip()

    return d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Makes the 'regular' ISIS functions act like the *_k functions*.

If you prefer the simplified argument structure and behavior of the
*_k functions*, but are constantly forgetting to put the ``_k`` on the end
of your function names, this module is for you.

For example, here's the 'regular' and *_k function* way of calling the ISIS
getkey program::

    import kalasiris as isis

    cube_file = 'some.cub'

    keyval = isis.getkey(cube_file, grpname='Instrument',
                         keyword='InstrumentId').stdout.strip()

    k_keyval = isis.getkey_k(cube_file, 'Instrument', 'InstrumentId')

And the values of ``keyval`` and ``k_keyval`` are the same.

However, if you do this::

    import kalasiris as isis

    cube_file = 'some.cub'

    key = isis.getkey(cube_file, 'Instrument', 'InstrumentId')

You'll get this exception::

    IndexError: only accepts 1 non-keyword argument to be from=

Because you tried to call ``isis.getkey()`` with the argument signature
of ``isis.getkey_k()`` and then ``isis.getkey()`` couldn't deal with it.

If you are always doing this, then instead of the above, do this::

    import kalasiris.sweetened as isis

    cube_file = 'some.cub'

    key = isis.getkey(cube_file, 'Instrument', 'InstrumentId')

And now the 'regularly' named functions will work the way that you expect the
*_k function* versions to act.
"""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from sys import modules

from .__init__ import *  # noqa F401,F403

from kalasiris import k_funcs

for k_func in dir(k_funcs):
    if k_func.endswith('_k'):
        reg_func = k_func.rsplit('_', maxsplit=1)[0]
        setattr(modules[__name__], reg_func, getattr(k_funcs, k_func))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Makes the 'regular' ISIS functions return what pysis did.

If you have old code that was built with pysis and is doing this::

    from pysis import isis

    value = isis.getkey(from_='W1467351325_4.map.cal.cub',
                        keyword='minimumringradius',
                        grp='mapping')

And you want to start using kalasiris, but don't want to overhaul
a bunch of code, you can just change the import line like so::

    import kalasiris.pysis as isis

    value = isis.getkey(from_='W1467351325_4.map.cal.cub',
                        keyword='minimumringradius',
                        grp='mapping')

And you should be good to go.  Note that this works for calls to
ISIS programs, but does not provide the pysis IsisPool functionality,
nor any of the non-ISIS pysis functions or classes, like ``pysis.cubefile``,
``pysis.specialpixels``, etc.
"""

# Copyright 2015, William Trevor Olson
# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import subprocess
import sys

import kalasiris as kala
from .kalasiris import _get_isis_program_names as gipn


class IsisException(Exception):
    """Base exception for pysis errors."""


class ProcessError(IsisException):
    """This exception is raised when an ISIS process returns a non-zero
    exit status."""

    def __init__(self, returncode, cmd, stdout, stderr):
        self.returncode = returncode
        self.cmd = cmd
        self.stdout = stdout
        self.stderr = stderr

        msg = (f'Command {self.cmd} returned non-zero '
               f'exit status {self.returncode}')
        super(ProcessError, self).__init__(msg)


def _build_pysis_fn(fn_name: str):
    """This factory builds a simple function call to wrap
    kalasiris function calls for pysis return types and
    pysis Exception types."""

    # Define the structure:
    def pysis_fn(*args, **kwargs):
        __name__ = fn_name                # noqa: F841
        __doc__ = f'Runs ISIS {fn_name}'  # noqa: F841

        try:
            kala_fn = getattr(kala, fn_name)
            return(kala_fn(*args, **kwargs).stdout.encode())
        except subprocess.CalledProcessError as err:
            raise ProcessError(err.returncode, err.cmd, err.stdout, err.stderr)

    # Then add it by name to the enclosing module, pysis.
    setattr(sys.modules[__name__], fn_name, pysis_fn)


for p in gipn():
    _build_pysis_fn(p)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `version` module."""

# Copyright 2019-2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import datetime
import unittest

import kalasiris.version as version


class TestVersion(unittest.TestCase):

    def test_get_from_string(self):
        s = [('''3.7.0        # Public version number
2019-04-30   # Release date
stable       # release stage (alpha, beta, stable)''',
              version.ISISversion(3, 7, 0, 'stable',
                                  datetime.date(2019, 4, 30))),
             ('''3.6.0        # Public version number
10-26-2018   # Release date
stable         # release stage (alpha, beta, stable)''',
              version.ISISversion(3, 6, 0, 'stable',
                                  datetime.date(2018, 10, 26))),
             ('''3.5.2.0
2018-04-06   # Version date
v007         # 3rd party libraries version
stable       # release stage (alpha, beta, stable)''',
              version.ISISversion(3, 5, 2, 'stable',
                                  datetime.date(2018, 4, 6)))]

        for v in s:
            with self.subTest(v=v[1]):
                self.assertEqual(v[1], version.get_from_string(v[0]))

    def test_fail_from_string(self):
        s = 'No version here.'
        self.assertRaises(ValueError, version.get_from_string, s)

    @unittest.skip("Not a robust test, only occaisional.")
    def test_version_info(self):
        # Who knows what version of ISIS will be loaded, this
        # is just for occaisionally testing this functionality.
        print(version.version_info())
